@@ -1,14 +1,15 @@
 import os
+import sys
 from copy import deepcopy
 
 from hatchling.metadata.utils import get_normalized_dependency, is_valid_project_name, normalize_project_name
 from hatchling.utils.constants import DEFAULT_CONFIG_FILE
 from hatchling.utils.fs import locate_file
 
-try:
+if sys.version_info >= (3, 11):
     import tomllib
-except ImportError:
-    import tomli as tomllib  # type: ignore
+else:
+    import tomli as tomllib
 
 
 def load_toml(path):
@@ -26,6 +27,9 @@ class ProjectMetadata:
         self._core = None
         self._hatch = None
 
+        self._core_raw_metadata = None
+        self._dynamic = None
+        self._name = None
         self._version = None
         self._project_file = None
 
@@ -46,6 +50,49 @@ class ProjectMetadata:
 
         return self._context
 
+    @property
+    def core_raw_metadata(self):
+        if self._core_raw_metadata is None:
+            if 'project' not in self.config:
+                raise ValueError('Missing `project` metadata table in configuration')
+
+            core_raw_metadata = self.config['project']
+            if not isinstance(core_raw_metadata, dict):
+                raise TypeError('The `project` configuration must be a table')
+
+            self._core_raw_metadata = core_raw_metadata
+
+        return self._core_raw_metadata
+
+    @property
+    def dynamic(self):
+        # Keep track of the original dynamic fields before depopulation
+        if self._dynamic is None:
+            dynamic = self.core_raw_metadata.get('dynamic', [])
+            if not isinstance(dynamic, list):
+                raise TypeError('Field `project.dynamic` must be an array')
+
+            for i, field in enumerate(dynamic, 1):
+                if not isinstance(field, str):
+                    raise TypeError(f'Field #{i} of field `project.dynamic` must be a string')
+
+            self._dynamic = list(dynamic)
+
+        return self._dynamic
+
+    @property
+    def name(self):
+        # Duplicate the name parsing here for situations where it's
+        # needed but metadata plugins might not be available
+        if self._name is None:
+            name = self.core_raw_metadata.get('name', '')
+            if not name:
+                raise ValueError('Missing required field `project.name`')
+
+            self._name = normalize_project_name(name)
+
+        return self._name
+
     @property
     def version(self):
         """
@@ -53,6 +100,8 @@ class ProjectMetadata:
         """
         if self._version is None:
             self._set_version()
+            if 'version' in self.dynamic and 'version' in self.core_raw_metadata['dynamic']:
+                self.core_raw_metadata['dynamic'].remove('version')
 
         return self._version
 
@@ -82,26 +131,23 @@ class ProjectMetadata:
     @property
     def core(self):
         if self._core is None:
-            if 'project' not in self.config:
-                raise ValueError('Missing `project` metadata table in configuration')
+            metadata = CoreMetadata(self.root, self.core_raw_metadata, self.hatch.metadata, self.context)
 
-            core_metadata = self.config['project']
-            if not isinstance(core_metadata, dict):
-                raise TypeError('The `project` configuration must be a table')
-
-            metadata = CoreMetadata(self.root, core_metadata, self.hatch.metadata, self.context)
+            # Save the fields
+            _ = self.dynamic
 
             metadata_hooks = self.hatch.metadata.hooks
             if metadata_hooks:
-                static_fields = set(core_metadata)
+                static_fields = set(self.core_raw_metadata)
                 if 'version' in self.hatch.config:
                     self._set_version(metadata)
-                    core_metadata['version'] = self.version
+                    self.core_raw_metadata['version'] = self.version
 
                 for metadata_hook in metadata_hooks.values():
-                    metadata_hook.update(core_metadata)
+                    metadata_hook.update(self.core_raw_metadata)
+                    metadata.add_known_classifiers(metadata_hook.get_known_classifiers())
 
-                new_fields = set(core_metadata) - static_fields
+                new_fields = set(self.core_raw_metadata) - static_fields
                 for new_field in new_fields:
                     if new_field in metadata.dynamic:
                         metadata.dynamic.remove(new_field)
@@ -148,6 +194,7 @@ class ProjectMetadata:
         if version is None:
             version = self.hatch.version.cached
             source = f'source `{self.hatch.version.source_name}`'
+            core_metadata._version_set = True
         else:
             source = 'field `project.version`'
 
@@ -269,6 +316,7 @@ class CoreMetadata:
         self._maintainers_data = None
         self._keywords = None
         self._classifiers = None
+        self._extra_classifiers = set()
         self._urls = None
         self._scripts = None
         self._gui_scripts = None
@@ -279,6 +327,9 @@ class CoreMetadata:
         self._optional_dependencies = None
         self._dynamic = None
 
+        # Indicates that the version has been successfully set dynamically
+        self._version_set = False
+
     @property
     def raw_name(self):
         """
@@ -324,7 +375,7 @@ class CoreMetadata:
         """
         if self._version is None:
             if 'version' not in self.config:
-                if 'version' not in self.dynamic:
+                if not self._version_set and 'version' not in self.dynamic:
                     raise ValueError(
                         'Field `project.version` can only be resolved dynamically '
                         'if `version` is in field `project.dynamic`'
@@ -535,7 +586,11 @@ class CoreMetadata:
             elif isinstance(data, str):
                 from hatchling.licenses.parse import normalize_license_expression
 
-                self._license_expression = normalize_license_expression(data)
+                try:
+                    self._license_expression = normalize_license_expression(data)
+                except ValueError as e:
+                    raise ValueError(f'Error parsing field `project.license` - {e}') from None
+
                 self._license = ''
             elif isinstance(data, dict):
                 if 'file' in data and 'text' in data:
@@ -799,6 +854,10 @@ class CoreMetadata:
         https://peps.python.org/pep-0621/#classifiers
         """
         if self._classifiers is None:
+            import bisect
+
+            from hatchling.metadata.classifiers import KNOWN_CLASSIFIERS, SORTED_CLASSIFIERS, is_private
+
             if 'classifiers' in self.config:
                 classifiers = self.config['classifiers']
                 if 'classifiers' in self.dynamic:
@@ -812,15 +871,24 @@ class CoreMetadata:
             if not isinstance(classifiers, list):
                 raise TypeError('Field `project.classifiers` must be an array')
 
+            known_classifiers = KNOWN_CLASSIFIERS | self._extra_classifiers
             unique_classifiers = set()
 
             for i, classifier in enumerate(classifiers, 1):
                 if not isinstance(classifier, str):
                     raise TypeError(f'Classifier #{i} of field `project.classifiers` must be a string')
+                elif not is_private(classifier) and classifier not in known_classifiers:
+                    raise ValueError(f'Unknown classifier in field `project.classifiers`: {classifier}')
 
                 unique_classifiers.add(classifier)
 
-            self._classifiers = sorted(unique_classifiers)
+            sorted_classifiers = list(SORTED_CLASSIFIERS)
+            for classifier in sorted(self._extra_classifiers - KNOWN_CLASSIFIERS):
+                bisect.insort(sorted_classifiers, classifier)
+
+            self._classifiers = sorted(
+                unique_classifiers, key=lambda c: -1 if is_private(c) else sorted_classifiers.index(c)
+            )
 
         return self._classifiers
 
@@ -999,7 +1067,8 @@ class CoreMetadata:
                 else:
                     if requirement.url and not self.hatch_metadata.allow_direct_references:
                         raise ValueError(
-                            f'Dependency #{i} of field `project.dependencies` cannot be a direct reference'
+                            f'Dependency #{i} of field `project.dependencies` cannot be a direct reference unless '
+                            f'field `tool.hatch.metadata.allow-direct-references` is set to `true`'
                         )
 
                     dependencies_complex[get_normalized_dependency(requirement)] = requirement
@@ -1074,12 +1143,16 @@ class CoreMetadata:
                         if requirement.url and not self.hatch_metadata.allow_direct_references:
                             raise ValueError(
                                 f'Dependency #{i} of option `{option}` of field `project.optional-dependencies` '
-                                f'cannot be a direct reference'
+                                f'cannot be a direct reference unless field '
+                                f'`tool.hatch.metadata.allow-direct-references` is set to `true`'
                             )
 
                         entries[get_normalized_dependency(requirement)] = requirement
 
-                normalized_option = normalize_project_name(option)
+                if self.hatch_metadata.allow_ambiguous_features:
+                    normalized_option = option
+                else:
+                    normalized_option = normalize_project_name(option)
                 if normalized_option in normalized_options:
                     raise ValueError(
                         f'Optional dependency groups `{normalized_options[normalized_option]}` and `{option}` of '
@@ -1111,18 +1184,13 @@ class CoreMetadata:
         https://peps.python.org/pep-0621/#dynamic
         """
         if self._dynamic is None:
-            dynamic = self.config.get('dynamic', [])
-            if not isinstance(dynamic, list):
-                raise TypeError('Field `project.dynamic` must be an array')
-
-            for i, field in enumerate(dynamic, 1):
-                if not isinstance(field, str):
-                    raise TypeError(f'Field #{i} of field `project.dynamic` must be a string')
-
-            self._dynamic = dynamic
+            self._dynamic = self.config.get('dynamic', [])
 
         return self._dynamic
 
+    def add_known_classifiers(self, classifiers):
+        self._extra_classifiers.update(classifiers)
+
     def validate_fields(self):
         # Trigger validation for everything
         for attribute in dir(self):
@@ -1280,6 +1348,7 @@ class HatchMetadataSettings:
         self.plugin_manager = plugin_manager
 
         self._allow_direct_references = None
+        self._allow_ambiguous_features = None
         self._hook_config = None
         self._hooks = None
 
@@ -1294,6 +1363,18 @@ class HatchMetadataSettings:
 
         return self._allow_direct_references
 
+    @property
+    def allow_ambiguous_features(self):
+        # TODO: remove in the first minor release after Jan 1, 2024
+        if self._allow_ambiguous_features is None:
+            allow_ambiguous_features = self.config.get('allow-ambiguous-features', False)
+            if not isinstance(allow_ambiguous_features, bool):
+                raise TypeError('Field `tool.hatch.metadata.allow-ambiguous-features` must be a boolean')
+
+            self._allow_ambiguous_features = allow_ambiguous_features
+
+        return self._allow_ambiguous_features
+
     @property
     def hook_config(self):
         if self._hook_config is None:
